/**
 * Reads in data from the ADC and makes it available to a Microblaze processor
 * over a FSL. Bits 24-31 give the channel number and bits 0-23 give the actual
 * data. Although the module implements a FSL slave as required, it discards
 * data silently.
 */
module adc_data_reader(clk, reset,
                       fsl_s_clk, fsl_s_read, fsl_s_data, fsl_s_control, fsl_s_exists,
                       fsl_m_clk, fsl_m_write, fsl_m_data, fsl_m_control, fsl_m_full,
                       gpio_clk, gpio_lat, gpio_data, gpio_channel
                       );

    input clk, reset;
    output fsl_s_clk;
    output fsl_s_read;
    input [31:0] fsl_s_data;
    input fsl_s_control;
    input fsl_s_exists;
    output fsl_m_clk;
    output fsl_m_write;
    output [31:0] fsl_m_data;
    output fsl_m_control;
    input fsl_m_full;
    input gpio_clk, gpio_lat;
    input gpio_data;
    input [1:0] gpio_channel;

    reg ack;

    wire en0, en1, en2, en3;
    wire gpio_lat_sync;
    wire [1:0] gpio_channel_sync;
    wire [11:0] c0, c1, c2, c3;
    reg [11:0] cdata;

    assign en0 = (gpio_channel == 2'd0);
    assign en1 = (gpio_channel == 2'd1);
    assign en2 = (gpio_channel == 2'd2);
    assign en3 = (gpio_channel == 2'd3);
    assign fsl_s_read = fsl_s_exists;
    assign fsl_m_control = 1'b0;
    assign fsl_m_write = gpio_lat_sync && !fsl_m_full && !ack;
    assign fsl_m_data = {6'b0, gpio_channel_sync, 12'b0, cdata};

    synchronizer gpio_lat_synchronizer(clk, reset, gpio_lat, gpio_lat_sync);
    synchronizer #(2) gpio_channel_synchronizer(clk, reset, gpio_channel, gpio_channel_sync);

    shiftreg #(12) sr0(gpio_clk, en0, gpio_data, c0);
    shiftreg #(12) sr1(gpio_clk, en1, gpio_data, c1);
    shiftreg #(12) sr2(gpio_clk, en2, gpio_data, c2);
    shiftreg #(12) sr3(gpio_clk, en3, gpio_data, c3);

    always @(*) begin
        case (gpio_channel_sync)
            2'd0: cdata <= c0;
            2'd1: cdata <= c1;
            2'd2: cdata <= c2;
            2'd3: cdata <= c3;
        endcase
    end

    initial begin
        ack <= 1'b0;
    end

    always @(posedge clk) begin
        if (reset) begin
            ack <= 1'b0;
        end
        else begin
            ack <= fsl_m_write ? 1'b1 : (gpio_lat_sync ? ack : 1'b0);
        end
    end

endmodule

/**
 * Simple serial-in (MSB to LSB), parallel-out shift register.
 */
module shiftreg #(parameter N=8) (clk, en, x, y);

    input clk, en;
    input x;
    output [N-1:0] y;

    reg [N-1:0] y;

    always @(posedge clk)
        if (en)
            y <= {x, y[N-1:1]};

endmodule

/**
 * Sign extender.
 */
module sign_extend #(parameter INLEN=16, OUTLEN=32) (x, y);

    input [INLEN-1:0] x;
    output [OUTLEN-1:0] y;

    assign y[INLEN-1:0] = x;

    generate
        genvar i;
        for (i = INLEN; i < OUTLEN; i = i + 1) begin:extend
            assign y[i] = x[INLEN-1];
        end
    endgenerate

endmodule

/**
 * N parallel 2-flop synchronizers. Resets low. clk should be driven by the
 * output side.
 */
module synchronizer #(parameter N=1) (clk, reset, x, y);

    input clk, reset;
    input [N-1:0] x;
    output [N-1:0] y;

    reg [N-1:0] y;
    reg [N-1:0] delay;

    initial begin
        y <= {N{1'b0}};
        delay <= {N{1'b0}};
    end

    always @(posedge clk) begin
        if (reset) begin
            y <= {N{1'b0}};
            delay <= {N{1'b0}};
        end
        else begin
            y <= delay;
            delay <= x;
        end
    end

endmodule

