/******************************************************************************
 * bioefilt.c - Bioelectricity filtering
 ******************************************************************************/

#include "bioefilt.h"

/* All coefficients are Q1.14 */

/* 60Hz comb filter coefficients */
const int FILTER_60HZ_NUM = 0x00003e60;
const int FILTER_60HZ_DEN = 0xffffc340;

/* EMG filter coefficients */
const int FILTER_EMG_SCALE[FILTER_EMG_NSEC] = {
    0x00003b6f,
};
const int FILTER_EMG_NUM[FILTER_EMG_NSEC][3] = {
    { 0x00004000, 0xffff8000, 0x00004000 },
};
const int FILTER_EMG_DEN[FILTER_EMG_NSEC][3] = {
    { 0x00004000, 0xffff8976, 0x00003731 },
};

/**
 * Convert a Qx.11 fixed point number (x >= 4, arbitrary upper bits) to Q4.11
 * (with correct sign extension).
 */
static int fix_q4_11(register int x) {
    if (x & 0x00010000) {
		x |= 0xffff8000;
	} else {
		x &= 0x00007fff;
	}
	return x;
}

/**
 * Implements an IIR filter. (num[0] and den[0] are expected to be 1.0.)
 */
#define FILTER_IIR(fn, pre) int fn(int x, int *delays) {\
    int i, j;\
    int w;\
    x <<= 14;\
    for (i = 0; i < pre##_NSEC; i++) {\
        j = i << 1;\
        w = x - pre##_DEN[i][1]*delays[j] - pre##_DEN[i][2]*delays[j+1];\
        x = w + pre##_NUM[i][1]*delays[j] + pre##_NUM[i][2]*delays[j+1];\
        x = fix_q4_11(x >> 14) * pre##_SCALE[i];\
        delays[j+1] = delays[j];\
        delays[j] = fix_q4_11(w >> 14);\
    }\
    return fix_q4_11(x >> 14);\
}

/**
 * Implements a comb filter.
 */
#define FILTER_COMB(fn, pre) int fn(int x, int *delays) {\
    int i;\
    int w;\
    w = (x << 14) - pre##_DEN*fix_q4_11(delays[pre##_ORDER-1] >> 14);\
    x = fix_q4_11((w - delays[pre##_ORDER-1]) >> 14) * pre##_NUM;\
    for (i = pre##_ORDER-1; i > 0; i--) {\
        delays[i] = delays[i-1];\
    }\
    delays[0] = w;\
    return fix_q4_11(x >> 14);\
}

FILTER_COMB(filter_60hz, FILTER_60HZ)

FILTER_IIR(filter_emg, FILTER_EMG)
